#ifndef LIBRAPID_INCLUDE
#define LIBRAPID_INCLUDE

#include <iostream>
#include <string>
#include <iomanip>

#include <librapid/config.hpp>

#include <librapid/utils/utils.hpp>
#include <librapid/test/librapid_test.hpp>
#include <librapid/stringmethods/format_number.hpp>
#include <librapid/math/rapid_math.hpp>
#include <librapid/autocast/autocast.hpp>
#include <librapid/array/multiarray.hpp>

// #if defined(__clang__)
// #pragma GCC diagnostic pop
// #elif defined(__GNUC__) || defined(__GNUG__)
// #pragma GCC diagnostic pop
// #elif defined(_MSC_VER)
// #pragma warning( pop )
// #endif

#endif // LIBRAPID_INCLUDE